/*
   client/playerview.cc
   This file is part of the Osirion project and is distributed under
   the terms and conditions of the GNU General Public License version 2
*/

#include <string>
#include <sstream>

#include "core/info.h"
#include "core/application.h"
#include "audio/audio.h"
#include "client/playerview.h"
#include "ui/ui.h"

namespace client
{

PlayerView::PlayerView(ui::Widget *parent) : ui::Widget(parent)
{
	set_label("playerview");
	set_border(false);
	set_background(false);

	view_notify = new Notifications(this);
	view_hud = new HUD(this);

	label_viewname = new ui::Label(this);
	label_viewname->set_label("viewnamelabel");
	label_viewname->set_alignment(ui::AlignCenter);
	label_viewname->set_background(true);
	label_viewname->set_font(ui::root()->font_large());

	view_chat = new Chat(this);
	view_map = new Map(this);
	view_entitymenu = new EntityMenu(this);
	view_buymenu = new BuyMenu(this);
	view_trademenu = new TradeMenu(this);
	
	//view_hud->set_focus();
	//view_hud->raise();
	//view_entitymenu->raise();
}

PlayerView::~PlayerView()
{
}

void PlayerView::clear()
{
	view_chat->clear();
	view_notify->clear();

	view_chat->hide();
	view_map->hide();
	view_entitymenu->hide();
	view_buymenu->hide();
	view_trademenu->hide();
}

void PlayerView::event_text(const std::string & text)
{
	view_chat->event_text(text);
	view_notify->event_text(text);
}

void PlayerView::toggle_map()
{

	if (!map()->visible()) {
		if (chat()->visible() && !chat()->small_view())
			chat()->hide();

		if (view_entitymenu->visible())
			view_entitymenu->hide();
	}

	map()->toggle();
	audio::play("ui/menu");

	if (map()->visible() && chat()->visible() && chat()->small_view()) {
		chat()->raise();
	}
}

void PlayerView::toggle_chat()
{
	if (!chat()->visible()) {
		if (map()->visible())
			map()->hide();

		if (view_entitymenu->visible())
			view_entitymenu->hide();
	}
	chat()->set_small_view(false);
	chat()->toggle();
	audio::play("ui/menu");
}

void PlayerView::toggle_chatbar()
{
	chat()->set_small_view(true);
	chat()->toggle();
}

void PlayerView::show_menu(const std::string & args)
{
	if (!args.size())
		return;

	if (!core::localplayer()->view())
		return;

	if (!core::localplayer()->view()->menus().size())
		return;

	std::stringstream argstr(args);
	std::string label;
	if (!(argstr >> label))
		return;

	if (label.compare("buy") == 0) {
		// buy menu, single item
		unsigned long id;

		if (argstr >> id) {
			// hide other menus
			view_entitymenu->hide();
			view_trademenu->hide();
						
			// requesting the info through game makes sure it is retreived from the server
			view_buymenu->set_item( core::game()->request_info(id)); 
			// show buy menu
			view_buymenu->show();			
		} else {
			con_print << "usage: view buy [info id] show the buy menu for this kind of item" << std::endl;
		}

	} else if (label.compare("trade") == 0) {
		// invetory based trade
		std::string typestr;
		
		if(argstr >> typestr) {
			aux::to_label(typestr);
		
			// hide other menus
			view_buymenu->hide();
			view_entitymenu->hide();
			// show trade menu			
			view_trademenu->set_itemtype(core::InfoType::find(typestr));
			view_trademenu->show();
		} else {
			con_print << "usage: view trade [string] show the trade menu for this type of items" << std::endl;
		}
		
	} else if (label.compare("hide") == 0) {
		// hide all menus
		view_buymenu->hide();
		view_entitymenu->hide();
		view_trademenu->hide();
	} else {
		// hide other menus
		view_buymenu->hide();
		view_trademenu->hide();
		// show other menus
		view_entitymenu->generate(core::localplayer()->view(), label.c_str());
		view_entitymenu->show();
	}

	if (chat()->visible() && chat()->small_view())
		chat()->raise();
}

void PlayerView::resize()
{
	const float smallmargin = ui::UI::elementsize.height();

	set_size(parent()->size());

	// reposition buy menu
	view_buymenu->event_resize();
	// reposition trade menu
	view_trademenu->event_resize();

	// set hud geometry
	view_hud->set_geometry(0, 0, width(), height());
	view_hud->event_resize();

	// reposition map
	view_map->set_size(width() -  smallmargin * 2, height() -  smallmargin * 4);
	view_map->set_location(smallmargin, smallmargin * 2);

	// reposition notifications
	view_notify->set_geometry(view_map->location(), view_map->size());

	// reposition labels
	label_viewname->set_size(ui::UI::elementsize.width() * 1.5f, ui::UI::elementsize.height());
	label_viewname->set_location(width() - label_viewname->width() - smallmargin, height() - label_viewname->height() - smallmargin * 0.5f);
}

void PlayerView::draw()
{
	const float smallmargin = ui::UI::elementsize.height();

	if (core::localplayer()->view()) {

		// hide hide when a view is set
		view_hud->hide();

		// docking view
		if (core::localplayer()->view()->menus().size()) {

			// entity with menus

			if (map()->visible()) {
				label_viewname->set_text(core::localplayer()->zone()->name());
			} else {
				label_viewname->set_text(core::localplayer()->view()->name());
			}

			label_viewname->show();

			if (view_entitymenu->generated_entity() != core::localplayer()->view()) {
				// initially show the menu
				show_menu("main");
				map()->hide();
				chat()->hide();
				audio::play("ui/menu");

			} else if (!view_entitymenu->visible() && !view_buymenu->visible() && !view_trademenu->visible() &&
					!map()->visible() && (!chat()->visible() || chat()->small_view())) {

				// show the menu if there's no other window open
				view_entitymenu->show();
				audio::play("ui/menu");
			}

			view_notify->set_size(width() - smallmargin * 3.0f - ui::UI::elementsize.width() * 1.5f, height() -  smallmargin * 4.0f);
			view_notify->set_location(smallmargin * 2.0f + ui::UI::elementsize.width() * 1.5f, smallmargin * 2.0f);

		} else {
			// entity without menus, plain view
			this->hide();
			return;
		}

	} else {
		view_notify->set_geometry(view_map->location(), view_map->size());

		if (view_entitymenu->visible()) {
			view_entitymenu->hide();
		}

		if (view_entitymenu->generated_entity()) {
			view_entitymenu->generate(0, 0);
		}

		if (view_buymenu->visible()) {
			view_buymenu->hide();
		}

		if (view_trademenu->visible()) {
			view_trademenu->hide();
		}
		
		if (map()->visible()) {
			label_viewname->set_text(core::localplayer()->zone()->name());
			label_viewname->show();
		} else {
			label_viewname->hide();
		}

		if (!map()->visible() && !chat()->visible()) {
			view_hud->set_focus();
		}

		view_hud->show();
	}

	// reposition chat widget
	if (view_chat->small_view()) {
		view_chat->set_size(width() - smallmargin * 2, font()->height() * 2);
		view_chat->set_location(smallmargin, height() - smallmargin *2 - view_chat->height());
	} else {
		view_chat->set_geometry(view_map->location(), view_map->size());
	}
	view_chat->event_resize();
}

}
